package net.sergeych.crypto2

import kotlinx.serialization.Serializable

@Serializable
class UniversalRing(
    val keyWithTags: Map<UniversalKey, Set<String>>,
) {
    constructor(vararg keys: UniversalKey) : this(keys.associateWith { setOf() })

    constructor(vararg keyTags: Pair<UniversalKey, String>)
            : this(keyTags.associate { it.first to setOf(it.second) })

    val decryptingKeys: Set<DecryptingKey> by lazy { keys<DecryptingKey>() }
    
    /**
     * Select keys of the specified type
     */
    inline fun <reified T> keys(): Set<T> =
        allKeys.mapNotNull { it as? T }.toSet()

    val allKeys: Set<UniversalKey> by lazy { keyWithTags.keys }
    
    inline fun <reified T> findKey(id: KeyId): UniversalKey? =
        allKeys.find { it is T && it.id == id }

    fun keysById(id: KeyId): List<UniversalKey> = allKeys.filter { it.id == id }

    /**
     * Return sequence of keys that have at least one of the [tags]
     */
    fun keysByTags(vararg tags: String) = sequence {
        for (e in keyWithTags.entries) {
            if (tags.any { it in e.value }) yield(e.key)
        }
    }

    /**
     * Get the first key of the specified type having the [tag]
     */
    inline fun <reified T> keyByTag(tag: String) = keysByTags(tag).first { it is T }

    @Suppress("unused")
    inline fun <reified T> keyByAnyTag(vararg tags: String) = keysByTags(*tags).first { it is T }

    /**
     * Get all keys with a given id. Note that _matching keys_ have the same id, see [KeyId] for more.
     */
    operator fun get(keyId: KeyId): Collection<UniversalKey> = allKeys.filter { it.id == keyId }

    fun getTags(key: UniversalKey): Set<String>? = keyWithTags[key]

    operator fun contains(element: UniversalKey): Boolean = element in keyWithTags

    /**
     * Add a key if not already in a ring, otherwise return existing ring.
     */
    operator fun plus(key: UniversalKey): UniversalRing =
        if (key in this) this else UniversalRing(keyWithTags + (key to setOf()))

    /**
     * Add a key and tags to the ring. If the key exists, add the tag to it.
     * See also [addTags].
     */
    operator fun plus(keyTag: Pair<UniversalKey, String>): UniversalRing =
        if (keyTag.first in this)
            addTags(keyTag.first, keyTag.second)
        else add(keyTag.first, keyTag.second)

    /**
     * Merge two rings. The result will contain all keys and all tags from
     * both rings.
     */
    operator fun plus(other: UniversalRing): UniversalRing {
        var result = keyWithTags.toMutableMap()
        for (e in other.keyWithTags.entries) {
            result[e.key]?.let {
                result[e.key] = it + e.value
            } ?: run {
                result[e.key] = e.value
            }
        }
        return UniversalRing(result)
    }

    /**
     * Add key and tags to the ring. If the key already exists, tags are merged.
     */
    fun add(key: UniversalKey, vararg tags: String): UniversalRing =
        UniversalRing(keyWithTags + (key to tags.toSet()))

    /**
     * Add key and tags to the ring. If the key already exists, tags are merged.
     */
    fun add(key: UniversalKey, tags: Collection<String>): UniversalRing =
        UniversalRing(keyWithTags + (key to tags.toSet()))

    /**
     * Add tags to the key, and add key if it is not in the ring.
     */
    fun addTags(key: UniversalKey, tags: Set<String>): UniversalRing {
        val kt1 = keyWithTags.toMutableMap()
        kt1[key]?.let {
            kt1[key] = it + tags
        } ?: run {
            kt1[key] = tags
        }
        return UniversalRing(kt1)
    }

    /**
     * Add tags to the key, and add key if it is not in the ring.
     */
    fun addTags(key: UniversalKey, vararg tags: String): UniversalRing =
        addTags(key, tags.toSet())

    /**
     * return the ring without [key] if existed, otherwise returns `this`.
     */
    operator fun minus(key: UniversalKey): UniversalRing =
        if (key in this) UniversalRing(keyWithTags.filter { it.key != key }) else this

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (other !is UniversalRing) return false
        return keyWithTags == other.keyWithTags
    }

    override fun hashCode(): Int {
        return keyWithTags.hashCode()
    }

    override fun toString(): String {
        val ss = keyWithTags.entries
            .joinToString(",") { "${it.value.joinToString { ":" }}:${it.key}" }
        return "Kr[$ss]"
    }

    infix fun equalKeys(other: UniversalRing): Boolean = allKeys == other.allKeys

    /**
     * Create a collection where no [tags] are specified for the [key]. If the key is not
     * in the collection, it returns this.
     */
    fun removeTags(key: UniversalKey, vararg tags: String): UniversalRing = removeTags(key, tags.toSet())

    /**
     * Create a collection where no [tags] are specified for the [key]. If the key is not
     * in the collection, it returns this.
     */
    fun removeTags(key: UniversalKey, tags: Set<String>): UniversalRing =
        keyWithTags[key]?.let { existingTags ->
            val kt1 = keyWithTags.toMutableMap()
            kt1[key] = existingTags - tags
            UniversalRing(kt1)
        } ?: this

    /**
     * Create string "report" of the ring contents. Note it has no trailing `\n`
     */
    fun ls(): String {
        val result = mutableListOf<String>()
        for( e in keyWithTags.entries) {
            result += "${e.key} ${e.value.joinToString(" ")}"
        }
        return result.joinToString("\n")
    }

    companion object {
        val EMPTY = UniversalRing(keyWithTags = emptyMap())

        /**
         * Join a collection of keyrings together (same as reducing with `+`). Correctly
         * works if there is no keyring (returns [EMPTY]), or only one keyring (returns
         * it with no extra work).
         *
         * Note that for the known number of rings using [plus] is more clear and expressive
         */
        fun join(keyRings: Collection<UniversalRing>): UniversalRing {
            if (keyRings.isEmpty()) return EMPTY
            if (keyRings.size == 1) return keyRings.first()
            return keyRings.reduce { l, r -> l + r }
        }

        fun from(vararg keys: DecryptingKey): UniversalRing =
            UniversalRing(keys.associate { UniversalKey.from(it) to setOf() } )
    }

}
